# http://github.com/jim-easterbrook/Photini
# IPTC "controlled vocabulary" data downloaded from http://cv.iptc.org/newscodes

# ©IPTC, International Press Telecommunications Council - https://iptc.org
# Date: 2019-12-06T12:00:00+00:00
# Licence: http://creativecommons.org/licenses/by/4.0/
# http://cv.iptc.org/newscodes/imageregiontype/
image_region_types = \
({'definition': {'en-GB': 'A living organism different from humans or flora'},
  'name': {'en-GB': 'animal'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/animal'},
 {'definition': {'en-GB': 'Artistic work'},
  'name': {'en-GB': 'artwork'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/artwork'},
 {'definition': {'en-GB': 'A line expressing a visual division of the image, '
                          'such as a horizon'},
  'name': {'en-GB': 'dividing line'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/dividingLine'},
 {'definition': {'en-GB': 'A living organism different from humans and '
                          'animals'},
  'name': {'en-GB': 'plant'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/plant'},
 {'definition': {'en-GB': 'A named area on the surface of the planet earth'},
  'name': {'en-GB': 'geographic area'},
  'note': {'en-GB': 'Specific details of the area can be expressed by other '
                    'metadata'},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/geoArea'},
 {'definition': {'en-GB': 'A graphic representation of information'},
  'name': {'en-GB': 'graphic'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/graphic'},
 {'definition': {'en-GB': 'Optical label such as barcode or QR code'},
  'name': {'en-GB': 'machine-readable code'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/machineCode'},
 {'definition': {'en-GB': 'A human being'},
  'name': {'en-GB': 'human'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/human'},
 {'definition': {'en-GB': 'A thing that was produced and can be handed over'},
  'name': {'en-GB': 'product'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/product'},
 {'definition': {'en-GB': 'Human readable script of any language'},
  'name': {'en-GB': 'text'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/text'},
 {'definition': {'en-GB': 'A structure with walls and roof in most cases'},
  'name': {'en-GB': 'building'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/building'},
 {'definition': {'en-GB': 'An object used for transporting something, like '
                          'car, train, ship, plane or bike'},
  'name': {'en-GB': 'vehicle'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/vehicle'},
 {'definition': {'en-GB': 'Substances providing nutrition for a living body'},
  'name': {'en-GB': 'food'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/food'},
 {'definition': {'en-GB': 'Something worn to cover the body'},
  'name': {'en-GB': 'clothing'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/clothing'},
 {'definition': {'en-GB': 'A special formation of stone mass'},
  'name': {'en-GB': 'rock formation'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/rockFormation'},
 {'definition': {'en-GB': 'A significant accumulation of water'},
  'name': {'en-GB': 'body of water'},
  'note': {'en-GB': 'Including a waterfall, a geyser and other phenomena of '
                    'water'},
  'uri': 'http://cv.iptc.org/newscodes/imageregiontype/bodyOfWater'})

# ©IPTC, International Press Telecommunications Council - https://iptc.org
# Date: 2019-12-06T12:00:00+00:00
# Licence: http://creativecommons.org/licenses/by/4.0/
# http://cv.iptc.org/newscodes/imageregionrole/
image_region_roles = \
({'definition': {'en-GB': 'Image region can be used for any cropping'},
  'name': {'en-GB': 'cropping'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/cropping'},
 {'definition': {'en-GB': 'Image region is recommended for cropping'},
  'name': {'en-GB': 'recommended cropping'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/recomCropping'},
 {'definition': {'en-GB': 'Image region suggested for cropping in landscape '
                          'format'},
  'name': {'en-GB': 'landscape format cropping'},
  'note': {'en-GB': 'Use for images of non-landscape format'},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/landscapeCropping'},
 {'definition': {'en-GB': 'Image region suggested for cropping in portrait '
                          'format'},
  'name': {'en-GB': 'portrait format cropping'},
  'note': {'en-GB': 'Use for images of non-portrait format'},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/portraitCropping'},
 {'definition': {'en-GB': 'Image region suggested for cropping in square '
                          'format'},
  'name': {'en-GB': 'square format cropping'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/squareCropping'},
 {'definition': {'en-GB': 'Image region of an item in a composite image'},
  'name': {'en-GB': 'composite image item'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/compositeImageItem'},
 {'definition': {'en-GB': 'Image region with a copyright different from the '
                          'copyright of the whole picture'},
  'name': {'en-GB': 'copyright region'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/copyrightRegion'},
 {'definition': {'en-GB': 'Image region contains a subject in the overall '
                          'scene.'},
  'name': {'en-GB': 'subject area'},
  'note': {'en-GB': 'Multiple regions of an image may be set as subject area.'},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/subjectArea'},
 {'definition': {'en-GB': 'Image region contains the main subject in the '
                          'overall scene. Same as the Exif SubjectArea.'},
  'name': {'en-GB': 'main subject area'},
  'note': {'en-GB': 'Only a single region of an image may be set as main '
                    'subject area.'},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/mainSubjectArea'},
 {'definition': {'en-GB': 'Image region contains a thing of special interest '
                          'to the viewer'},
  'name': {'en-GB': 'area of interest'},
  'note': {},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/areaOfInterest'},
 {'definition': {'en-GB': 'Image region is dedicated to a specific business '
                          'use'},
  'name': {'en-GB': 'business use'},
  'note': {'en-GB': 'In addition a more granular role could be expressed by a '
                    'term of a role CV of this business.'},
  'uri': 'http://cv.iptc.org/newscodes/imageregionrole/businessUse'})

